@@ -1,7 +1,7 @@
 /**************************************************************************
  *   files.c  --  This file is part of GNU nano.                          *
  *                                                                        *
- *   Copyright (C) 1999-2011, 2013-2021 Free Software Foundation, Inc.    *
+ *   Copyright (C) 1999-2011, 2013-2022 Free Software Foundation, Inc.    *
  *   Copyright (C) 2015-2021 Benno Schulenberg                            *
  *                                                                        *
  *   GNU nano is free software: you can redistribute it and/or modify     *
@@ -33,6 +33,10 @@
 
 #define RW_FOR_ALL  (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH)
 
+#ifndef HAVE_FSYNC
+# define fsync(...)  0
+#endif
+
 /* Add an item to the circular list of openfile structs. */
 void make_new_buffer(void)
 {
@@ -143,7 +147,7 @@ const char *locking_suffix = ".swp";
  * existing version of that file.  Return TRUE on success; FALSE otherwise. */
 bool write_lockfile(const char *lockfilename, const char *filename, bool modified)
 {
-#ifdef HAVE_PWD_H
+#if defined(HAVE_PWD_H) && defined(HAVE_GETEUID)
 	pid_t mypid = getpid();
 	uid_t myuid = geteuid();
 	struct passwd *mypwuid = getpwuid(myuid);
@@ -303,14 +307,14 @@ char *do_lockfile(const char *filename, bool ask_the_user)
 		free(postedname);
 		free(pidstring);
 
-		choice = do_yesno_prompt(FALSE, promptstr);
+		choice = ask_user(YESORNO, promptstr);
 		free(promptstr);
 
 		/* When the user cancelled while we're still starting up, quit. */
-		if (choice < 0 && !we_are_running)
+		if (choice == CANCEL && !we_are_running)
 			finish();
 
-		if (choice < 1) {
+		if (choice != YES) {
 			free(lockfilename);
 			wipe_statusbar();
 			return SKIPTHISFILE;
@@ -346,8 +350,18 @@ bool has_valid_path(const char *filename)
 	char *parentdir = dirname(namecopy);
 	struct stat parentinfo;
 	bool validity = FALSE;
+	bool gone = FALSE;
 
-	if (stat(parentdir, &parentinfo) == -1) {
+	if (strcmp(parentdir, ".") == 0) {
+		char *currentdir = realpath(".", NULL);
+
+		gone = (currentdir == NULL && errno == ENOENT);
+		free(currentdir);
+	}
+
+	if (gone)
+		statusline(ALERT, _("The working directory has disappeared"));
+	else if (stat(parentdir, &parentinfo) == -1) {
 		if (errno == ENOENT)
 			/* TRANSLATORS: Keep the next ten messages at most 76 characters. */
 			statusline(ALERT, _("Directory '%s' does not exist"), parentdir);
@@ -411,7 +425,7 @@ bool open_buffer(const char *filename, bool new_one)
 			free(realname);
 			return FALSE;
 		}
-#else
+#elif defined(HAVE_GETEUID)
 		if (new_one && !(fileinfo.st_mode & (S_IWUSR|S_IWGRP|S_IWOTH)) &&
 						geteuid() == ROOT_UID)
 			statusline(ALERT, _("%s is meant to be read-only"), realname);
@@ -560,13 +574,14 @@ void redecorate_after_switch(void)
 	/* Prevent a possible Shift selection from getting cancelled. */
 	shift_held = TRUE;
 
+	/* If the switched-to buffer gave an error during opening, show the message
+	 * once; otherwise, indicate on the status bar which file we switched to. */
 	if (openfile->errormessage) {
 		statusline(ALERT, openfile->errormessage);
 		free(openfile->errormessage);
 		openfile->errormessage = NULL;
 	} else
-	/* Indicate on the status bar where we switched to. */
-	mention_name_and_linecount();
+		mention_name_and_linecount();
 }
 
 /* Switch to the previous entry in the circular list of buffers. */
@@ -675,9 +690,13 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 	block_sigwinch(TRUE);
 #endif
 
+#ifdef HAVE_FLOCKFILE
 	/* Lock the file before starting to read it, to avoid the overhead
 	 * of locking it for each single byte that we read from it. */
 	flockfile(f);
+#else
+# define getc_unlocked  getc
+#endif
 
 	control_C_was_pressed = FALSE;
 
@@ -741,8 +760,10 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 
 	errornumber = errno;
 
+#ifdef HAVE_FUNLOCKFILE
 	/* We are done with the file, unlock it. */
 	funlockfile(f);
+#endif
 
 #ifndef NANO_TINY
 	block_sigwinch(FALSE);
@@ -811,25 +832,29 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 #ifndef NANO_TINY
 	else if (format == MAC_FILE)
 		/* TRANSLATORS: Keep the next three messages at most 78 characters. */
-		statusline(HUSH, P_("Read %zu line (Converted from Mac format)",
+		statusline(REMARK, P_("Read %zu line (Converted from Mac format)",
 						"Read %zu lines (Converted from Mac format)",
 						num_lines), num_lines);
 	else if (format == DOS_FILE)
-		statusline(HUSH, P_("Read %zu line (Converted from DOS format)",
+		statusline(REMARK, P_("Read %zu line (Converted from DOS format)",
 						"Read %zu lines (Converted from DOS format)",
 						num_lines), num_lines);
 #endif
-	else
-		statusline(HUSH, P_("Read %zu line", "Read %zu lines",
+	else if ((!ISSET(MINIBAR) && !ISSET(ZERO)) || (we_are_running && undoable))
+		statusline(REMARK, P_("Read %zu line", "Read %zu lines",
 						num_lines), num_lines);
 
+	report_size = TRUE;
+
 	/* If we inserted less than a screenful, don't center the cursor. */
-	if (undoable && less_than_a_screenful(was_lineno, was_leftedge))
+	if (undoable && less_than_a_screenful(was_lineno, was_leftedge)) {
 		focusing = FALSE;
 #ifdef ENABLE_COLOR
-	else if (undoable)
-		precalc_multicolorinfo();
+		perturbed = TRUE;
+	} else if (undoable) {
+		recook = TRUE;
 #endif
+	}
 
 #ifndef NANO_TINY
 	if (undoable)
@@ -898,7 +923,7 @@ int open_file(const char *filename, bool new_one, FILE **f)
 			statusline(ALERT, _("Error reading %s: %s"), filename, strerror(errno));
 			close(fd);
 			fd = -1;
-		} else
+		} else if (!ISSET(ZERO) || we_are_running)
 			statusbar(_("Reading..."));
 	}
 
@@ -948,7 +973,9 @@ static pid_t pid_of_command = -1;
 /* Send an unconditional kill signal to the running external command. */
 void cancel_the_command(int signal)
 {
+#ifdef SIGKILL
 	kill(pid_of_command, SIGKILL);
+#endif
 }
 
 /* Send the text that starts at the given line to file descriptor fd. */
@@ -971,10 +998,12 @@ void send_data(const linestruct *line, int fd)
 /* Execute the given command in a shell.  Return TRUE on success. */
 bool execute_command(const char *command)
 {
+#if defined(HAVE_FORK) && defined(HAVE_PIPE) && defined(HAVE_WAIT)
 	int from_fd[2], to_fd[2];
 		/* The pipes through which text will be written and read. */
 	struct sigaction oldaction, newaction = {{0}};
 		/* Original and temporary handlers for SIGINT. */
+	ssize_t was_lineno = (openfile->mark ? 0 : openfile->current->lineno);
 	const bool should_pipe = (command[0] == '|');
 	FILE *stream;
 
@@ -1006,6 +1035,7 @@ bool execute_command(const char *command)
 		if (should_pipe) {
 			if (dup2(to_fd[0], STDIN_FILENO) < 0)
 				exit(5);
+			close(from_fd[1]);
 			close(to_fd[1]);
 		}
 
@@ -1052,7 +1082,8 @@ bool execute_command(const char *command)
 			}
 			add_undo(CUT, NULL);
 			do_snip(openfile->mark != NULL, openfile->mark == NULL, FALSE);
-			openfile->filetop->has_anchor = FALSE;
+			if (openfile->filetop->next == NULL)
+				openfile->filetop->has_anchor = FALSE;
 			update_undo(CUT);
 		}
 
@@ -1088,8 +1119,11 @@ bool execute_command(const char *command)
 	else
 		read_file(stream, 0, "pipe", TRUE);
 
-	if (should_pipe && !ISSET(MULTIBUFFER))
+	if (should_pipe && !ISSET(MULTIBUFFER)) {
+		if (was_lineno)
+			goto_line_posx(was_lineno, 0);
 		add_undo(COUPLE_END, N_("filtering"));
+	}
 
 	/* Wait for the external command (and possibly data sender) to terminate. */
 	wait(NULL);
@@ -1104,12 +1138,16 @@ bool execute_command(const char *command)
 	terminal_init();
 
 	return TRUE;
+#else
+	return FALSE;
+#endif
 }
 #endif /* NANO_TINY */
 
-/* Insert a file into the current buffer, or into a new buffer when
- * the MULTIBUFFER flag is set. */
-void do_insertfile(bool execute)
+/* Insert a file into the current buffer (or into a new buffer).  But when
+ * execute is TRUE, run a command in the shell and insert its output into
+ * the buffer, or just run one of the tools listed in the help lines. */
+void insert_a_file_or(bool execute)
 {
 	int response;
 	const char *msg;
@@ -1158,15 +1196,9 @@ void do_insertfile(bool execute)
 
 		present_path = mallocstrcpy(present_path, "./");
 
-		response = do_prompt(
-#ifndef NANO_TINY
-							execute ? MEXECUTE :
-#endif
-							MINSERTFILE, given,
-#ifndef NANO_TINY
-							execute ? &execute_history :
-#endif
-							NULL, edit_refresh, msg,
+		response = do_prompt(execute ? MEXECUTE : MINSERTFILE, given,
+							execute ? &execute_history : NULL,
+							edit_refresh, msg,
 #ifdef ENABLE_OPERATINGDIR
 							operating_dir != NULL ? operating_dir :
 #endif
@@ -1242,7 +1274,7 @@ void do_insertfile(bool execute)
 				if (*answer != '\0') {
 					execute_command(answer);
 #ifdef ENABLE_HISTORIES
-					update_history(&execute_history, answer);
+					update_history(&execute_history, answer, PRUNE_DUPLICATE);
 #endif
 				}
 
@@ -1275,7 +1307,7 @@ void do_insertfile(bool execute)
 					if (!execute)
 #endif
 					if (has_old_position(answer, &priorline, &priorcol))
-						do_gotolinecolumn(priorline, priorcol, FALSE, FALSE);
+						goto_line_and_column(priorline, priorcol, FALSE, FALSE);
 				}
 #endif
 				/* Update title bar and color info for this new buffer. */
@@ -1306,10 +1338,10 @@ void do_insertfile(bool execute)
 }
 
 /* If the current mode of operation allows it, go insert a file. */
-void do_insertfile_void(void)
+void do_insertfile(void)
 {
 	if (!in_restricted_mode())
-		do_insertfile(FALSE);
+		insert_a_file_or(FALSE);
 }
 
 #ifndef NANO_TINY
@@ -1317,7 +1349,7 @@ void do_insertfile_void(void)
 void do_execute(void)
 {
 	if (!in_restricted_mode())
-		do_insertfile(TRUE);
+		insert_a_file_or(TRUE);
 }
 #endif
 
@@ -1325,106 +1357,37 @@ void do_execute(void)
  * absolute path (plus filename) when the path exists, and NULL when not. */
 char *get_full_path(const char *origpath)
 {
-	char *allocation, *here, *target, *last_slash;
-	char *just_filename = NULL;
-	int attempts = 0;
+	char *untilded, *target, *slash;
 	struct stat fileinfo;
-	bool path_only;
 
 	if (origpath == NULL)
 		return NULL;
 
-	allocation = nmalloc(PATH_MAX + 1);
-	here = getcwd(allocation, PATH_MAX + 1);
+	untilded = real_dir_from_tilde(origpath);
+	target = realpath(untilded, NULL);
+	slash = strrchr(untilded, '/');
 
-	/* If getting the current directory failed, go up one level and try again,
-	 * until we find an existing directory, and use that as the current one. */
-	while (here == NULL && attempts < 20) {
-		IGNORE_CALL_RESULT(chdir(".."));
-		here = getcwd(allocation, PATH_MAX + 1);
-		attempts++;
-	}
+	/* If realpath() returned NULL, try without the last component,
+	 * as this can be a file that does not exist yet. */
+	if (!target && slash && slash[1]) {
+		*slash = '\0';
+		target = realpath(untilded, NULL);
 
-	/* If we found a directory, make sure its path ends in a slash. */
-	if (here != NULL) {
-		if (strcmp(here, "/") != 0) {
-			here = nrealloc(here, strlen(here) + 2);
-			strcat(here, "/");
-		}
-	} else {
-		here = copy_of("");
-		free(allocation);
-	}
-
-	target = real_dir_from_tilde(origpath);
-
-	/* Determine whether the target path refers to a directory.  If statting
-	 * target fails, however, assume that it refers to a new, unsaved buffer. */
-	path_only = (stat(target, &fileinfo) != -1 && S_ISDIR(fileinfo.st_mode));
-
-	/* If the target is a directory, make sure its path ends in a slash. */
-	if (path_only) {
-		size_t length = strlen(target);
-
-		if (target[length - 1] != '/') {
-			target = nrealloc(target, length + 2);
-			strcat(target, "/");
+		/* Upon success, re-add the last component of the original path. */
+		if (target) {
+			target = nrealloc(target, strlen(target) + strlen(slash + 1) + 1);
+			strcat(target, slash + 1);
 		}
 	}
 
-	last_slash = strrchr(target, '/');
-
-	/* If the target path does not contain a slash, then it is a bare filename
-	 * and must therefore be located in the working directory. */
-	if (last_slash == NULL) {
-		just_filename = target;
-		target = here;
-	} else {
-		/* If target contains a filename, separate the two. */
-		if (!path_only) {
-			just_filename = copy_of(last_slash + 1);
-			*(last_slash + 1) = '\0';
-		}
-
-		/* If we can't change to the target directory, give up.  Otherwise,
-		 * get the canonical path to this target directory. */
-		if (chdir(target) == -1) {
-			free(target);
-			target = NULL;
-		} else {
-			free(target);
-
-			allocation = nmalloc(PATH_MAX + 1);
-			target = getcwd(allocation, PATH_MAX + 1);
-
-			/* If we got a result, make sure it ends in a slash.
-			 * Otherwise, ensure that we return NULL. */
-			if (target != NULL) {
-				if (strcmp(target, "/") != 0) {
-					target = nrealloc(target, strlen(target) + 2);
-					strcat(target, "/");
-				}
-			} else {
-				path_only = TRUE;
-				free(allocation);
-			}
-
-			/* Finally, go back to where we were before.  We don't check
-			 * for an error, since we can't do anything if we get one. */
-			IGNORE_CALL_RESULT(chdir(here));
-		}
-
-		free(here);
+	/* Ensure that a non-apex directory path ends with a slash. */
+	if (target && target[1] && stat(target, &fileinfo) == 0 &&
+								S_ISDIR(fileinfo.st_mode)) {
+		target = nrealloc(target, strlen(target) + 2);
+		strcat(target, "/");
 	}
 
-	/* If we were given more than a bare path, concatenate the target path
-	 * with the filename portion to get the full, absolute file path. */
-	if (!path_only && target != NULL) {
-		target = nrealloc(target, strlen(target) + strlen(just_filename) + 1);
-		strcat(target, just_filename);
-	}
-
-	free(just_filename);
+	free(untilded);
 
 	return target;
 }
@@ -1656,6 +1619,7 @@ bool make_backup_of(char *realname)
 	if (backup_file == NULL)
 		goto problem;
 
+#ifdef HAVE_FCHOWN
 	/* Try to change owner and group to those of the original file;
 	 * ignore permission errors, as a normal user cannot change the owner. */
 	if (fchown(descriptor, openfile->statinfo->st_uid,
@@ -1663,7 +1627,8 @@ bool make_backup_of(char *realname)
 		fclose(backup_file);
 		goto problem;
 	}
-
+#endif
+#ifdef HAVE_FCHMOD
 	/* Set the backup's permissions to those of the original file.
 	 * It is not a security issue if this fails, as we have created
 	 * the file with just read and write permission for the owner. */
@@ -1671,6 +1636,7 @@ bool make_backup_of(char *realname)
 		fclose(backup_file);
 		goto problem;
 	}
+#endif
 
 	original = fopen(realname, "rb");
 
@@ -1735,8 +1701,8 @@ bool make_backup_of(char *realname)
 	 * ask the user what to do, because if something goes wrong during the
 	 * save of the file itself, its contents may be lost. */
 	/* TRANSLATORS: Try to keep this message at most 76 characters. */
-	if (errno != ENOSPC && do_yesno_prompt(FALSE, _("Cannot make backup; "
-								"continue and save actual file? ")) == 1)
+	if (errno != ENOSPC && ask_user(YESORNO, _("Cannot make backup; "
+							"continue and save actual file? ")) == YES)
 		return TRUE;
 
 	/* TRANSLATORS: The %s is the reason of failure. */
@@ -1977,6 +1943,19 @@ bool write_file(const char *name, FILE *thefile, bool normal,
 		statusline(ALERT, _("Error writing %s: %s"), realname, strerror(errno));
 
   cleanup_and_exit:
+#ifndef NANO_TINY
+		if (errno == ENOSPC && normal) {
+			napms(3200); lastmessage = VACUUM;
+			/* TRANSLATORS: This warns for data loss when the disk is full. */
+			statusline(ALERT, _("File on disk has been truncated!"));
+			napms(3200); lastmessage = VACUUM;
+			/* TRANSLATORS: This is a suggestion to the user,
+			 * where "resume" means resuming from suspension.
+			 * Try to keep this at most 76 characters. */
+			statusline(ALERT, _("Maybe ^T^Z, make room on disk, resume, then ^S^X"));
+			stat_with_alloc(realname, &openfile->statinfo);
+		}
+#endif
 		free(tempname);
 		free(realname);
 		return FALSE;
@@ -2030,7 +2009,7 @@ bool write_file(const char *name, FILE *thefile, bool normal,
 	}
 
 #ifndef NANO_TINY
-	if (ISSET(MINIBAR) && LINES > 1 && annotate)
+	if (ISSET(MINIBAR) && !ISSET(ZERO) && LINES > 1 && annotate)
 		report_size = TRUE;
 	else
 #endif
@@ -2058,7 +2037,7 @@ bool write_region_to_file(const char *name, FILE *stream, bool normal,
 	get_region(&topline, &top_x, &botline, &bot_x);
 
 	/* When needed, prepare a magic end line for the region. */
-	if (bot_x > 0 && !ISSET(NO_NEWLINES)) {
+	if (normal && bot_x > 0 && !ISSET(NO_NEWLINES)) {
 		stopper = make_new_node(botline);
 		stopper->data = copy_of("");
 	} else
@@ -2089,13 +2068,12 @@ bool write_region_to_file(const char *name, FILE *stream, bool normal,
 }
 #endif /* !NANO_TINY */
 
-/* Write the current buffer to disk.  If the mark is on, write the current
- * marked selection to disk.  If exiting is TRUE, write the entire buffer
- * to disk regardless of whether the mark is on.  Do not ask for a name
- * when withprompt is FALSE nor when the SAVE_ON_EXIT flag is set and the
- * buffer already has a name.  Return 0 on error, 1 on success, and 2 when
+/* Write the current buffer (or marked region) to disk.  If exiting is TRUE,
+ * write the entire buffer regardless of whether the mark is on.  Do not ask
+ * for a name when withprompt is FALSE (nor when doing save-on-exit and the
+ * buffer already has a name).  Return 0 on error, 1 on success, and 2 when
  * the buffer is to be discarded. */
-int do_writeout(bool exiting, bool withprompt)
+int write_it_out(bool exiting, bool withprompt)
 {
 	char *given;
 		/* The filename we offer, or what the user typed so far. */
@@ -2116,9 +2094,10 @@ int do_writeout(bool exiting, bool withprompt)
 #endif
 
 	while (TRUE) {
-		int response = 0, choice = 0;
 		functionptrtype func;
 		const char *msg;
+		int response = 0;
+		int choice = NO;
 #ifndef NANO_TINY
 		const char *formatstr = (openfile->fmt == DOS_FILE) ? _(" [DOS Format]") :
 						(openfile->fmt == MAC_FILE) ? _(" [Mac Format]") : "";
@@ -2184,22 +2163,24 @@ int do_writeout(bool exiting, bool withprompt)
 		} else
 #endif
 #ifndef NANO_TINY
-		if (func == dos_format_void) {
+		if (func == dos_format) {
 			openfile->fmt = (openfile->fmt == DOS_FILE) ? NIX_FILE : DOS_FILE;
 			continue;
-		} else if (func == mac_format_void) {
+		} else if (func == mac_format) {
 			openfile->fmt = (openfile->fmt == MAC_FILE) ? NIX_FILE : MAC_FILE;
 			continue;
-		} else if (func == backup_file_void) {
+		} else if (func == back_it_up) {
 			TOGGLE(MAKE_BACKUP);
 			continue;
-		} else if (func == prepend_void) {
-			method = (method == PREPEND) ? OVERWRITE : PREPEND;
+		} else if (func == prepend_it || func == append_it) {
+			if (func == prepend_it)
+				method = (method == PREPEND) ? OVERWRITE : PREPEND;
+			else
+				method = (method == APPEND) ? OVERWRITE : APPEND;
+			if (strcmp(answer, openfile->filename) == 0)
+				given[0] = '\0';
 			continue;
-		} else if (func == append_void) {
-			method = (method == APPEND) ? OVERWRITE : APPEND;
-			continue;
-		}
+		} else
 #endif
 		if (func == do_help)
 			continue;
@@ -2257,8 +2238,8 @@ int do_writeout(bool exiting, bool withprompt)
 					if (exiting || !openfile->mark)
 #endif
 					{
-						if (do_yesno_prompt(FALSE, _("Save file under "
-												"DIFFERENT NAME? ")) < 1)
+						if (ask_user(YESORNO, _("Save file under "
+												"DIFFERENT NAME? ")) != YES)
 							continue;
 						maychange = TRUE;
 					}
@@ -2271,12 +2252,12 @@ int do_writeout(bool exiting, bool withprompt)
 
 					sprintf(message, question, name);
 
-					choice = do_yesno_prompt(FALSE, message);
+					choice = ask_user(YESORNO, message);
 
 					free(message);
 					free(name);
 
-					if (choice < 1)
+					if (choice != YES)
 						continue;
 				}
 			}
@@ -2292,7 +2273,7 @@ int do_writeout(bool exiting, bool withprompt)
 				warn_and_briefly_pause(_("File on disk has changed"));
 
 				/* TRANSLATORS: Try to keep this at most 76 characters. */
-				choice = do_yesno_prompt(FALSE, _("File was modified "
+				choice = ask_user(YESORNO, _("File was modified "
 								"since you opened it; continue saving? "));
 				wipe_statusbar();
 
@@ -2300,14 +2281,16 @@ int do_writeout(bool exiting, bool withprompt)
 				 * overwrite the file right here when requested. */
 				if (ISSET(SAVE_ON_EXIT) && withprompt) {
 					free(given);
-					if (choice == 1)
+					if (choice == YES)
 						return write_file(openfile->filename, NULL,
 											NORMAL, OVERWRITE, NONOTES);
-					else if (choice == 0)
+					else if (choice == NO)  /* Discard buffer */
 						return 2;
 					else
 						return 0;
-				} else if (choice != 1) {
+				} else if (choice == CANCEL && exiting) {
+					continue;
+				} else if (choice != YES) {
 					free(given);
 					return 1;
 				}
@@ -2331,17 +2314,17 @@ int do_writeout(bool exiting, bool withprompt)
 }
 
 /* Write the current buffer to disk, or discard it. */
-void do_writeout_void(void)
+void do_writeout(void)
 {
 	/* If the user chose to discard the buffer, close it. */
-	if (do_writeout(FALSE, TRUE) == 2)
+	if (write_it_out(FALSE, TRUE) == 2)
 		close_and_go();
 }
 
 /* If it has a name, write the current buffer to disk without prompting. */
 void do_savefile(void)
 {
-	if (do_writeout(FALSE, FALSE) == 2)
+	if (write_it_out(FALSE, FALSE) == 2)
 		close_and_go();
 }
 
@@ -2418,13 +2401,13 @@ int diralphasort(const void *va, const void *vb)
 /* Return TRUE when the given path is a directory. */
 bool is_dir(const char *path)
 {
-	char *realpath = real_dir_from_tilde(path);
+	char *thepath = real_dir_from_tilde(path);
 	struct stat fileinfo;
 	bool retval;
 
-	retval = (stat(realpath, &fileinfo) != -1 && S_ISDIR(fileinfo.st_mode));
+	retval = (stat(thepath, &fileinfo) != -1 && S_ISDIR(fileinfo.st_mode));
 
-	free(realpath);
+	free(thepath);
 
 	return retval;
 }
@@ -2629,6 +2612,7 @@ char *input_tab(char *morsel, size_t *place, void (*refresh_func)(void), bool *l
 
 	/* If there is more than one possible completion, show a sorted list. */
 	if (num_matches > 1) {
+		int lastrow = editwinrows - 1 - (ISSET(ZERO) && LINES > 1 ? 1 : 0);
 		size_t longest_name = 0;
 		size_t nrows, ncols;
 		int row;
@@ -2654,7 +2638,7 @@ char *input_tab(char *morsel, size_t *place, void (*refresh_func)(void), bool *l
 		ncols = (COLS + 1) / (longest_name + 2);
 		nrows = (num_matches + ncols - 1) / ncols;
 
-		row = (nrows < editwinrows - 1) ? editwinrows - nrows - 1 : 0;
+		row = (nrows < lastrow) ? lastrow - nrows : 0;
 
 		/* Blank the edit window and hide the cursor. */
 		blank_edit();
@@ -2664,23 +2648,23 @@ char *input_tab(char *morsel, size_t *place, void (*refresh_func)(void), bool *l
 		for (match = 0; match < num_matches; match++) {
 			char *disp;
 
-			wmove(edit, row, (longest_name + 2) * (match % ncols));
+			wmove(midwin, row, (longest_name + 2) * (match % ncols));
 
-			if (row == editwinrows - 1 && (match + 1) % ncols == 0 &&
+			if (row == lastrow && (match + 1) % ncols == 0 &&
 											match + 1 < num_matches) {
-				waddstr(edit, _("(more)"));
+				waddstr(midwin, _("(more)"));
 				break;
 			}
 
 			disp = display_string(matches[match], 0, longest_name, FALSE, FALSE);
-			waddstr(edit, disp);
+			waddstr(midwin, disp);
 			free(disp);
 
 			if ((match + 1) % ncols == 0)
 				row++;
 		}
 
-		wnoutrefresh(edit);
+		wnoutrefresh(midwin);
 		*listed = TRUE;
 	}
 
