@@ -1,7 +1,7 @@
 /**************************************************************************
  *   search.c  --  This file is part of GNU nano.                         *
  *                                                                        *
- *   Copyright (C) 1999-2011, 2013-2021 Free Software Foundation, Inc.    *
+ *   Copyright (C) 1999-2011, 2013-2022 Free Software Foundation, Inc.    *
  *   Copyright (C) 2015-2020 Benno Schulenberg                            *
  *                                                                        *
  *   GNU nano is free software: you can redistribute it and/or modify     *
@@ -65,6 +65,9 @@ void tidy_up_after_search(void)
 	if (openfile->mark)
 		refresh_needed = TRUE;
 #endif
+#ifdef ENABLE_COLOR
+	recook |= perturbed;
+#endif
 }
 
 /* Prepare the prompt and ask the user what to search for.  Keep looping
@@ -118,7 +121,7 @@ void search_init(bool replacing, bool retain_answer)
 			if (*answer != '\0') {
 				last_search = mallocstrcpy(last_search, answer);
 #ifdef ENABLE_HISTORIES
-				update_history(&search_history, answer);
+				update_history(&search_history, answer, PRUNE_DUPLICATE);
 #endif
 			}
 
@@ -152,7 +155,7 @@ void search_init(bool replacing, bool retain_answer)
 			} else
 				replacing = !replacing;
 		} else if (func == flip_goto) {
-			do_gotolinecolumn(openfile->current->lineno,
+			goto_line_and_column(openfile->current->lineno,
 								openfile->placewewant + 1, TRUE, TRUE);
 			break;
 		} else
@@ -186,7 +189,7 @@ int findnextstr(const char *needle, bool whole_word_only, int modus,
 		/* The time we last looked at the keyboard. */
 
 	/* Set non-blocking input so that we can just peek for a Cancel. */
-	nodelay(edit, TRUE);
+	nodelay(midwin, TRUE);
 
 	if (begin == NULL)
 		came_full_circle = FALSE;
@@ -226,14 +229,14 @@ int findnextstr(const char *needle, bool whole_word_only, int modus,
 #ifndef NANO_TINY
 		if (the_window_resized) {
 			regenerate_screen();
-			nodelay(edit, TRUE);
+			nodelay(midwin, TRUE);
 			statusbar(_("Searching..."));
 			feedback = 1;
 		}
 #endif
 		/* If we're back at the beginning, then there is no needle. */
 		if (came_full_circle) {
-			nodelay(edit, FALSE);
+			nodelay(midwin, FALSE);
 			return 0;
 		}
 
@@ -244,7 +247,7 @@ int findnextstr(const char *needle, bool whole_word_only, int modus,
 		 * but stop when spell-checking or replacing in a region. */
 		if (line == NULL) {
 			if (whole_word_only || modus == INREGION) {
-				nodelay(edit, FALSE);
+				nodelay(midwin, FALSE);
 				return 0;
 			}
 
@@ -268,7 +271,7 @@ int findnextstr(const char *needle, bool whole_word_only, int modus,
 
 		/* Glance at the keyboard once every second, to check for a Cancel. */
 		if (time(NULL) - lastkbcheck > 0) {
-			int input = wgetch(edit);
+			int input = wgetch(midwin);
 
 			lastkbcheck = time(NULL);
 
@@ -276,7 +279,7 @@ int findnextstr(const char *needle, bool whole_word_only, int modus,
 			while (input != ERR) {
 				if (input == ESC_CODE) {
 					napms(20);
-					input = wgetch(edit);
+					input = wgetch(midwin);
 					meta_key = TRUE;
 				} else
 					meta_key = FALSE;
@@ -289,12 +292,12 @@ int findnextstr(const char *needle, bool whole_word_only, int modus,
 					statusbar(_("Cancelled"));
 					/* Clear out the key buffer (in case a macro is running). */
 					while (input != ERR)
-						input = parse_kbinput(NULL);
-					nodelay(edit, FALSE);
+						input = get_input(NULL);
+					nodelay(midwin, FALSE);
 					return -2;
 				}
 
-				input = wgetch(edit);
+				input = wgetch(midwin);
 			}
 
 			if (++feedback > 0)
@@ -306,7 +309,7 @@ int findnextstr(const char *needle, bool whole_word_only, int modus,
 
 	found_x = found - line->data;
 
-	nodelay(edit, FALSE);
+	nodelay(midwin, FALSE);
 
 	/* Ensure that the found occurrence is not beyond the starting x. */
 	if (came_full_circle && ((!ISSET(BACKWARDS_SEARCH) && (found_x > begin_x ||
@@ -327,9 +330,7 @@ int findnextstr(const char *needle, bool whole_word_only, int modus,
 		spotlighted = TRUE;
 		light_from_col = xplustabs();
 		light_to_col = wideness(line->data, found_x + found_len);
-		if (!ISSET(SHOW_CURSOR))
-			hide_cursor = TRUE;
-		edit_refresh();
+		refresh_needed = TRUE;
 	}
 #endif
 
@@ -445,8 +446,8 @@ void go_looking(void)
  * text in the passed string only when create is TRUE. */
 int replace_regexp(char *string, bool create)
 {
-	const char *c = answer;
 	size_t replacement_size = 0;
+	const char *c = answer;
 
 	/* Iterate through the replacement text to handle subexpression
 	 * replacement using \1, \2, \3, etc. */
@@ -524,11 +525,11 @@ char *replace_line(const char *needle)
 ssize_t do_replace_loop(const char *needle, bool whole_word_only,
 		const linestruct *real_current, size_t *real_current_x)
 {
+	bool skipone = ISSET(BACKWARDS_SEARCH);
+	bool replaceall = FALSE;
+	int modus = REPLACING;
 	ssize_t numreplaced = -1;
 	size_t match_len;
-	bool replaceall = FALSE;
-	bool skipone = ISSET(BACKWARDS_SEARCH);
-	int modus = REPLACING;
 #ifndef NANO_TINY
 	linestruct *was_mark = openfile->mark;
 	linestruct *top, *bot;
@@ -555,7 +556,7 @@ ssize_t do_replace_loop(const char *needle, bool whole_word_only,
 	came_full_circle = FALSE;
 
 	while (TRUE) {
-		int choice = 0;
+		int choice = NO;
 		int result = findnextstr(needle, whole_word_only, modus,
 						&match_len, skipone, real_current, *real_current_x);
 
@@ -591,21 +592,21 @@ ssize_t do_replace_loop(const char *needle, bool whole_word_only,
 			edit_refresh();
 
 			/* TRANSLATORS: This is a prompt. */
-			choice = do_yesno_prompt(TRUE, _("Replace this instance?"));
+			choice = ask_user(YESORALLORNO, _("Replace this instance?"));
 
 			spotlighted = FALSE;
 
-			if (choice == -1)  /* The replacing was cancelled. */
+			if (choice == CANCEL)
 				break;
-			else if (choice == 2)
-				replaceall = TRUE;
+
+			replaceall = (choice == ALL);
 
 			/* When "No" or moving backwards, the search routine should
 			 * first move one character further before continuing. */
 			skipone = (choice == 0 || ISSET(BACKWARDS_SEARCH));
 		}
 
-		if (choice == 1 || replaceall) {  /* Yes, replace it. */
+		if (choice == YES || replaceall) {
 			size_t length_change;
 			char *altered;
 
@@ -659,6 +660,10 @@ ssize_t do_replace_loop(const char *needle, bool whole_word_only,
 			free(openfile->current->data);
 			openfile->current->data = altered;
 
+#ifdef ENABLE_COLOR
+			check_the_multis(openfile->current);
+			refresh_needed = FALSE;
+#endif
 			set_modified();
 			as_an_at = TRUE;
 			numreplaced++;
@@ -706,7 +711,7 @@ void ask_for_and_do_replacements(void)
 #ifdef ENABLE_HISTORIES
 	/* When not "", add the replace string to the replace history list. */
 	if (response == 0)
-		update_history(&replace_history, answer);
+		update_history(&replace_history, answer, PRUNE_DUPLICATE);
 #endif
 
 	/* When cancelled, or when a function was run, get out. */
@@ -724,7 +729,7 @@ void ask_for_and_do_replacements(void)
 	openfile->current = beginline;
 	openfile->current_x = begin_x;
 
-	edit_refresh();
+	refresh_needed = TRUE;
 
 	if (numreplaced >= 0)
 		statusline(REMARK, P_("Replaced %zd occurrence",
@@ -734,6 +739,12 @@ void ask_for_and_do_replacements(void)
 /* Go to the specified line and x position. */
 void goto_line_posx(ssize_t line, size_t pos_x)
 {
+#ifdef ENABLE_COLOR
+	if (line > openfile->edittop->lineno + editwinrows ||
+				(ISSET(SOFTWRAP) && line > openfile->current->lineno))
+		recook |= perturbed;
+#endif
+
 	for (openfile->current = openfile->filetop; line > 1 &&
 				openfile->current != openfile->filebot; line--)
 		openfile->current = openfile->current->next;
@@ -747,8 +758,8 @@ void goto_line_posx(ssize_t line, size_t pos_x)
 /* Go to the specified line and column, or ask for them if interactive
  * is TRUE.  In the latter case also update the screen afterwards.
  * Note that both the line and column number should be one-based. */
-void do_gotolinecolumn(ssize_t line, ssize_t column, bool retain_answer,
-		bool interactive)
+void goto_line_and_column(ssize_t line, ssize_t column, bool retain_answer,
+							bool interactive)
 {
 	if (interactive) {
 		/* Ask for the line and column. */
@@ -792,6 +803,12 @@ void do_gotolinecolumn(ssize_t line, ssize_t column, bool retain_answer,
 	if (line < 1)
 		line = 1;
 
+#ifdef ENABLE_COLOR
+	if (line > openfile->edittop->lineno + editwinrows ||
+				(ISSET(SOFTWRAP) && line > openfile->current->lineno))
+		recook |= perturbed;
+#endif
+
 	/* Iterate to the requested line. */
 	for (openfile->current = openfile->filetop; line > 1 &&
 				openfile->current != openfile->filebot; line--)
@@ -844,9 +861,9 @@ void do_gotolinecolumn(ssize_t line, ssize_t column, bool retain_answer,
 }
 
 /* Go to the specified line and column, asking for them beforehand. */
-void do_gotolinecolumn_void(void)
+void do_gotolinecolumn(void)
 {
-	do_gotolinecolumn(openfile->current->lineno,
+	goto_line_and_column(openfile->current->lineno,
 						openfile->placewewant + 1, FALSE, TRUE);
 }
 
@@ -994,6 +1011,11 @@ void go_to_and_confirm(linestruct *line)
 	if (line != openfile->current) {
 		openfile->current = line;
 		openfile->current_x = 0;
+#ifdef ENABLE_COLOR
+		if (line->lineno > openfile->edittop->lineno + editwinrows ||
+					(ISSET(SOFTWRAP) && line->lineno > was_current->lineno))
+			recook |= perturbed;
+#endif
 		edit_redraw(was_current, CENTERING);
 		statusbar(_("Jumped to anchor"));
 	} else if (openfile->current->has_anchor)
